#!/usr/bin/env python

"""
SELFIES: a robust representation of semantically constrained graphs with an
         example application in chemistry.

SELFIES (SELF-referencIng Embedded Strings) is a general-purpose,
sequence-based, robust representation of semantically constrained graphs.
It is based on a Chomsky type-2 grammar, augmented with two self-referencing
functions. A main objective is to use SELFIES as direct input into machine
learning models, in particular in generative models, for the generation of
outputs with high validity.

The code presented here is a concrete application of SELFIES in chemistry, for
the robust representation of molecules.

    Typical usage example:
        import selfies as sf

        benzene = "C1=CC=CC=C1"
        benzene_selfies = sf.encoder(benzene)
        benzene_smiles = sf.decoder(benzene_selfies)

For comments, bug reports or feature ideas, please send an email to
mario.krenn@utoronto.ca and alan@aspuru.com.
"""

__version__ = "2.1.0"

__all__ = [
    "encoder",
    "decoder",
    "get_preset_constraints",
    "get_semantic_robust_alphabet",
    "get_semantic_constraints",
    "set_semantic_constraints",
    "len_selfies",
    "split_selfies",
    "get_alphabet_from_selfies",
    "selfies_to_encoding",
    "batch_selfies_to_flat_hot",
    "encoding_to_selfies",
    "batch_flat_hot_to_selfies",
    "EncoderError",
    "DecoderError"
]

from .bond_constraints import (
    get_preset_constraints,
    get_semantic_constraints,
    get_semantic_robust_alphabet,
    set_semantic_constraints
)
from .decoder import decoder
from .encoder import encoder
from .exceptions import DecoderError, EncoderError
from .utils.encoding_utils import (
    batch_flat_hot_to_selfies,
    batch_selfies_to_flat_hot,
    encoding_to_selfies,
    selfies_to_encoding
)
from .utils.selfies_utils import (
    get_alphabet_from_selfies,
    len_selfies,
    split_selfies
)



import functools
from itertools import product
from typing import Dict, Set, Union

from selfies.constants import ELEMENTS, INDEX_ALPHABET

_DEFAULT_CONSTRAINTS = {
    "H": 1, "F": 1, "Cl": 1, "Br": 1, "I": 1,
    "B": 3, "B+1": 2, "B-1": 4,
    "O": 2, "O+1": 3, "O-1": 1,
    "N": 3, "N+1": 4, "N-1": 2,
    "C": 4, "C+1": 5, "C-1": 3,
    "P": 5, "P+1": 6, "P-1": 4,
    "S": 6, "S+1": 7, "S-1": 5,
    "?": 8
}

_PRESET_CONSTRAINTS = {
    "default": dict(_DEFAULT_CONSTRAINTS),
    "octet_rule": dict(_DEFAULT_CONSTRAINTS),
    "hypervalent": dict(_DEFAULT_CONSTRAINTS)
}
_PRESET_CONSTRAINTS["octet_rule"].update(
    {"S": 2, "S+1": 3, "S-1": 1, "P": 3, "P+1": 4, "P-1": 2}
)
_PRESET_CONSTRAINTS["hypervalent"].update(
    {"Cl": 7, "Br": 7, "I": 7, "N": 5}
)

_current_constraints = _PRESET_CONSTRAINTS["default"]


def get_preset_constraints(name: str) -> Dict[str, int]:
    """Returns the preset semantic constraints with the given name.

    Besides the aforementioned default constraints, :mod:`selfies` offers
    other preset constraints for convenience; namely, constraints that
    enforce the `octet rule <https://en.wikipedia.org/wiki/Octet_rule>`_
    and constraints that accommodate `hypervalent molecules
    <https://en.wikipedia.org/wiki/Hypervalent_molecule>`_.

    The differences between these constraints can be summarized as follows:

    .. table::
        :align: center
        :widths: auto

        +-----------------+-----------+---+---+-----+-----+---+-----+-----+
        |                 | Cl, Br, I | N | P | P+1 | P-1 | S | S+1 | S-1 |
        +-----------------+-----------+---+---+-----+-----+---+-----+-----+
        | ``default``     |     1     | 3 | 5 |  6  |  4  | 6 |  7  |  5  |
        +-----------------+-----------+---+---+-----+-----+---+-----+-----+
        | ``octet_rule``  |     1     | 3 | 3 |  4  |  2  | 2 |  3  |  1  |
        +-----------------+-----------+---+---+-----+-----+---+-----+-----+
        | ``hypervalent`` |     7     | 5 | 5 |  6  |  4  | 6 |  7  |  5  |
        +-----------------+-----------+---+---+-----+-----+---+-----+-----+

    :param name: the preset name: ``default`` or ``octet_rule`` or
        ``hypervalent``.
    :return: the preset constraints with the specified name, represented
        as a dictionary which maps atoms (the keys) to their bonding capacities
        (the values).
    """

    if name not in _PRESET_CONSTRAINTS:
        raise ValueError("unrecognized preset name '{}'".format(name))
    return dict(_PRESET_CONSTRAINTS[name])


def get_semantic_constraints() -> Dict[str, int]:
    """Returns the semantic constraints that :mod:`selfies` is currently
    operating on.

    :return: the current semantic constraints, represented as a dictionary
        which maps atoms (the keys) to their bonding capacities (the values).
    """

    global _current_constraints
    return dict(_current_constraints)


def set_semantic_constraints(
        bond_constraints: Union[str, Dict[str, int]] = "default"
) -> None:
    """Updates the semantic constraints that :mod:`selfies` operates on.

    If the input is a string, the new constraints are taken to be
    the preset named ``bond_constraints``
    (see :func:`selfies.get_preset_constraints`).

    Otherwise, the input is a dictionary representing the new constraints.
    This dictionary maps atoms (the keys) to non-negative bonding
    capacities (the values); the atoms are specified by strings
    of the form ``E`` or ``E+C`` or ``E-C``,
    where ``E`` is an element symbol and ``C`` is a positive integer.
    For example, one may have:

       * ``bond_constraints["I-1"] = 0``
       * ``bond_constraints["C"] = 4``

    This dictionary must also contain the special ``?`` key, which indicates
    the bond capacities of all atoms that are not explicitly listed
    in the dictionary.

    :param bond_constraints: the name of a preset, or a dictionary
        representing the new semantic constraints.
    :return: ``None``.
    """

    global _current_constraints

    if isinstance(bond_constraints, str):
        _current_constraints = get_preset_constraints(bond_constraints)

    elif isinstance(bond_constraints, dict):

        # error checking
        if "?" not in bond_constraints:
            raise ValueError("bond_constraints missing '?' as a key")

        for key, value in bond_constraints.items():

            # error checking for keys
            j = max(key.find("+"), key.find("-"))
            if key == "?":
                valid = True
            elif j == -1:
                valid = (key in ELEMENTS)
            else:
                valid = (key[:j] in ELEMENTS) and key[j + 1:].isnumeric()
            if not valid:
                err_msg = "invalid key '{}' in bond_constraints".format(key)
                raise ValueError(err_msg)

            # error checking for values
            if not (isinstance(value, int) and value >= 0):
                err_msg = "invalid value at " \
                          "bond_constraints['{}'] = {}".format(key, value)
                raise ValueError(err_msg)

        _current_constraints = dict(bond_constraints)

    else:
        raise ValueError("bond_constraints must be a str or dict")

    # clear cache since we changed alphabet
    get_semantic_robust_alphabet.cache_clear()
    get_bonding_capacity.cache_clear()


@functools.lru_cache()
def get_semantic_robust_alphabet() -> Set[str]:
    """Returns a subset of all SELFIES symbols that are constrained
    by :mod:`selfies` under the current semantic constraints.

    :return: a subset of all SELFIES symbols that are semantically constrained.
    """

    alphabet_subset = set()
    bonds = {"": 1, "=": 2, "#": 3}

    # add atomic symbols
    for (a, c), (b, m) in product(_current_constraints.items(), bonds.items()):
        if (m > c) or (a == "?"):
            continue
        symbol = "[{}{}]".format(b, a)
        alphabet_subset.add(symbol)

    # add branch and ring symbols
    for i in range(1, 4):
        alphabet_subset.add("[Ring{}]".format(i))
        alphabet_subset.add("[=Ring{}]".format(i))
        alphabet_subset.add("[Branch{}]".format(i))
        alphabet_subset.add("[=Branch{}]".format(i))
        alphabet_subset.add("[#Branch{}]".format(i))

    alphabet_subset.update(INDEX_ALPHABET)

    return alphabet_subset


@functools.lru_cache()
def get_bonding_capacity(element: str, charge: int) -> int:
    """Returns the bonding capacity of a given atom, under the current
    semantic constraints.

    :param element: the element of the input atom.
    :param charge: the charge of the input atom.
    :return: the bonding capacity of the input atom.
    """

    key = element
    if charge != 0:
        key += "{:+}".format(charge)

    if key in _current_constraints:
        return _current_constraints[key]
    else:
        return _current_constraints["?"]



from selfies.utils.smiles_utils import atom_to_smiles, smiles_to_atom


def modernize_symbol(symbol):
    """Converts a SELFIES symbol from <v2 to its latest equivalent.

    :param symbol: an old SELFIES symbol.
    :return: the latest equivalent of the input symbol, or the input symbol
        itself, if no such equivalent exists.
    """

    if symbol in _SYMBOL_UPDATE_TABLE:
        return _SYMBOL_UPDATE_TABLE[symbol]

    if symbol[-5:] == "expl]":  # e.g. [XXXexpl]
        if symbol[1] in "=#/\\":
            bond_char, atom_symbol = symbol[1], symbol[2:-5]
        else:
            bond_char, atom_symbol = "", symbol[1:-5]

        atom = smiles_to_atom("[{}]".format(atom_symbol))
        if (atom is not None) and (not atom.is_aromatic):
            atom_symbol = atom_to_smiles(atom, brackets=False)  # standardize
            symbol = "[{}{}]".format(bond_char, atom_symbol)

    return symbol


def _build_update_table():
    update_table = dict()
    for L in range(1, 4):
        entries = [
            ("[Branch{}_1]", "[Branch{}]"),
            ("[Branch{}_2]", "[=Branch{}]"),
            ("[Branch{}_3]", "[#Branch{}]"),
            ("[Expl=Ring{}]", "[=Ring{}]"),
            ("[Expl#Ring{}]", "[#Ring{}]"),
            ("[Expl/Ring{}]", "[//Ring{}]"),
            ("[Expl\\Ring{}]", "[\\\\Ring{}]")
        ]

        for old, new in entries:
            update_table[old.format(L)] = new.format(L)
    return update_table


_SYMBOL_UPDATE_TABLE = _build_update_table()



import functools
import itertools
import re
from typing import Any, List, Optional, Tuple

from selfies.constants import (
    ELEMENTS,
    INDEX_ALPHABET,
    INDEX_CODE,
    ORGANIC_SUBSET
)
from selfies.mol_graph import Atom
from selfies.utils.smiles_utils import smiles_to_bond


def process_atom_symbol(symbol: str) -> Optional[Tuple[Any, Atom]]:
    try:
        output = _PROCESS_ATOM_CACHE[symbol]
    except KeyError:
        output = _process_atom_selfies_no_cache(symbol)
        if output is None:
            return None
        _PROCESS_ATOM_CACHE[symbol] = output

    bond_info, atom_fac = output
    atom = atom_fac()
    if atom.bonding_capacity < 0:
        return None  # too many Hs (e.g. [CH9]
    return bond_info, atom


def process_branch_symbol(symbol: str) -> Optional[Tuple[int, int]]:
    try:
        return _PROCESS_BRANCH_CACHE[symbol]
    except KeyError:
        return None


def process_ring_symbol(symbol: str) -> Optional[Tuple[int, int, Any]]:
    try:
        return _PROCESS_RING_CACHE[symbol]
    except KeyError:
        return None


def next_atom_state(
        bond_order: int, bond_cap: int, state: int
) -> Tuple[int, Optional[int]]:
    if state == 0:
        bond_order = 0

    bond_order = min(bond_order, state, bond_cap)
    bonds_left = bond_cap - bond_order
    next_state = None if (bonds_left == 0) else bonds_left
    return bond_order, next_state


def next_branch_state(
        branch_type: int, state: int
) -> Tuple[int, Optional[int]]:
    assert 1 <= branch_type <= 3
    assert state > 1

    branch_init_state = min(state - 1, branch_type)
    next_state = state - branch_init_state
    return branch_init_state, next_state


def next_ring_state(
        ring_type: int, state: int
) -> Tuple[int, Optional[int]]:
    assert state > 0

    bond_order = min(ring_type, state)
    bonds_left = state - bond_order
    next_state = None if (bonds_left == 0) else bonds_left
    return bond_order, next_state


def get_index_from_selfies(*symbols: List[str]) -> int:
    index = 0
    for i, c in enumerate(reversed(symbols)):
        index += INDEX_CODE.get(c, 0) * (len(INDEX_CODE) ** i)
    return index


def get_selfies_from_index(index: int) -> List[str]:
    if index < 0:
        raise IndexError()
    elif index == 0:
        return [INDEX_ALPHABET[0]]

    symbols = []
    base = len(INDEX_ALPHABET)
    while index:
        symbols.append(INDEX_ALPHABET[index % base])
        index //= base
    return symbols[::-1]


# =============================================================================
# Caches (for computational speed)
# =============================================================================


SELFIES_ATOM_PATTERN = re.compile(
    r"^[\[]"  # opening square bracket [
    r"([=#/\\]?)"  # bond char
    r"(\d*)"  # isotope number (optional, e.g. 123, 26)
    r"([A-Z][a-z]?)"  # element symbol
    r"([@]{0,2})"  # chiral_tag (optional, only @ and @@ supported)
    r"((?:[H]\d)?)"  # H count (optional, e.g. H1, H3)
    r"((?:[+-][1-9]+)?)"  # charge (optional, e.g. +1)
    r"[]]$"  # closing square bracket ]
)


def _process_atom_selfies_no_cache(symbol):
    m = SELFIES_ATOM_PATTERN.match(symbol)
    if m is None:
        return None
    bond_char, isotope, element, chirality, h_count, charge = m.groups()

    if symbol[1 + len(bond_char):-1] in ORGANIC_SUBSET:
        atom_fac = functools.partial(Atom, element=element, is_aromatic=False)
        return smiles_to_bond(bond_char), atom_fac

    isotope = None if (isotope == "") else int(isotope)
    if element not in ELEMENTS:
        return None
    chirality = None if (chirality == "") else chirality

    s = h_count
    if s == "":
        h_count = 0
    else:
        h_count = int(s[1:])

    s = charge
    if s == "":
        charge = 0
    else:
        charge = int(s[1:])
        charge *= 1 if (s[0] == "+") else -1

    atom_fac = functools.partial(
        Atom,
        element=element,
        is_aromatic=False,
        isotope=isotope,
        chirality=chirality,
        h_count=h_count,
        charge=charge
    )

    return smiles_to_bond(bond_char), atom_fac


def _build_atom_cache():
    cache = dict()
    common_symbols = [
        "[#C+1]", "[#C-1]", "[#C]", "[#N+1]", "[#N]", "[#O+1]", "[#P+1]",
        "[#P-1]", "[#P]", "[#S+1]", "[#S-1]", "[#S]", "[=C+1]", "[=C-1]",
        "[=C]", "[=N+1]", "[=N-1]", "[=N]", "[=O+1]", "[=O]", "[=P+1]",
        "[=P-1]", "[=P]", "[=S+1]", "[=S-1]", "[=S]", "[Br]", "[C+1]", "[C-1]",
        "[C]", "[Cl]", "[F]", "[H]", "[I]", "[N+1]", "[N-1]", "[N]", "[O+1]",
        "[O-1]", "[O]", "[P+1]", "[P-1]", "[P]", "[S+1]", "[S-1]", "[S]"
    ]

    for symbol in common_symbols:
        cache[symbol] = _process_atom_selfies_no_cache(symbol)
    return cache


def _build_branch_cache():
    cache = dict()
    for L in range(1, 4):
        for bond_char in ["", "=", "#"]:
            symbol = "[{}Branch{}]".format(bond_char, L)
            cache[symbol] = (smiles_to_bond(bond_char)[0], L)
    return cache


def _build_ring_cache():
    cache = dict()
    for L in range(1, 4):
        # [RingL], [=RingL], [#RingL]
        for bond_char in ["", "=", "#"]:
            symbol = "[{}Ring{}]".format(bond_char, L)
            order, stereo = smiles_to_bond(bond_char)
            cache[symbol] = (order, L, (stereo, stereo))

        # [-/RingL], [\/RingL], [\-RingL], ...
        for lchar, rchar in itertools.product(["-", "/", "\\"], repeat=2):
            if lchar == rchar == "-":
                continue
            symbol = "[{}{}Ring{}]".format(lchar, rchar, L)
            order, lstereo = smiles_to_bond(lchar)
            order, rstereo = smiles_to_bond(rchar)
            cache[symbol] = (order, L, (lstereo, rstereo))
    return cache


_PROCESS_ATOM_CACHE = _build_atom_cache()

_PROCESS_BRANCH_CACHE = _build_branch_cache()

_PROCESS_RING_CACHE = _build_ring_cache()



import functools
import itertools
from typing import List, Optional, Union
from dataclasses import dataclass, field

from selfies.bond_constraints import get_bonding_capacity
from selfies.constants import AROMATIC_VALENCES
from selfies.utils.matching_utils import find_perfect_matching


@dataclass
class Attribution:
    """A dataclass that contains token string and its index.
    """
    #: token index
    index: int
    #: token string
    token: str


@dataclass
class AttributionMap:
    """A mapping from input to single output token showing which
    input tokens created the output token.
    """
    #: Index of output token
    index: int
    #: Output token
    token: str
    #: List of input tokens that created the output token
    attribution: List[Attribution] = field(default_factory=list)


class Atom:
    """An atom with associated specifications (e.g. charge, chirality).
    """

    def __init__(
            self,
            element: str,
            is_aromatic: bool,
            isotope: Optional[int] = None,
            chirality: Optional[str] = None,
            h_count: Optional[int] = None,
            charge: int = 0
    ):
        self.index = None
        self.element = element
        self.is_aromatic = is_aromatic
        self.isotope = isotope
        self.chirality = chirality
        self.h_count = h_count
        self.charge = charge

    @property
    @functools.lru_cache()
    def bonding_capacity(self):
        bond_cap = get_bonding_capacity(self.element, self.charge)
        bond_cap -= 0 if (self.h_count is None) else self.h_count
        return bond_cap

    def invert_chirality(self) -> None:
        if self.chirality == "@":
            self.chirality = "@@"
        elif self.chirality == "@@":
            self.chirality = "@"


class DirectedBond:
    """A bond that contains directional information.
    """

    def __init__(
            self,
            src: int,
            dst: int,
            order: Union[int, float],
            stereo: Optional[str],
            ring_bond: bool
    ):
        self.src = src
        self.dst = dst
        self.order = order
        self.stereo = stereo
        self.ring_bond = ring_bond


class MolecularGraph:
    """A molecular graph.

    Molecules can be viewed as weighted undirected graphs. However, SMILES
    and SELFIES strings are more naturally represented as weighted directed
    graphs, where the direction of the edges specifies the order of atoms
    and bonds in the string.
    """

    def __init__(self, attributable=False):
        self._roots = list()  # stores root atoms, where traversal begins
        self._atoms = list()  # stores atoms in this graph
        self._bond_dict = dict()  # stores all bonds in this graph
        self._adj_list = list()  # adjacency list, representing this graph
        self._bond_counts = list()  # stores number of bonds an atom has made
        self._ring_bond_flags = list()  # stores if an atom makes a ring bond
        self._delocal_subgraph = dict()  # delocalization subgraph
        self._attribution = dict()  # attribution of each atom/bond
        self._attributable = attributable

    def __len__(self):
        return len(self._atoms)

    def has_bond(self, a: int, b: int) -> bool:
        if a > b:
            a, b = b, a
        return (a, b) in self._bond_dict

    def has_out_ring_bond(self, src: int) -> bool:
        return self._ring_bond_flags[src]

    def get_attribution(
        self,
        o: Union[DirectedBond, Atom]
    ) -> List[Attribution]:
        if self._attributable and o in self._attribution:
            return self._attribution[o]
        return None

    def get_roots(self) -> List[int]:
        return self._roots

    def get_atom(self, idx: int) -> Atom:
        return self._atoms[idx]

    def get_atoms(self) -> List[Atom]:
        return self._atoms

    def get_dirbond(self, src, dst) -> DirectedBond:
        return self._bond_dict[(src, dst)]

    def get_out_dirbonds(self, src: int) -> List[DirectedBond]:
        return self._adj_list[src]

    def get_bond_count(self, idx: int) -> int:
        return self._bond_counts[idx]

    def add_atom(self, atom: Atom, mark_root: bool = False) -> Atom:
        atom.index = len(self)

        if mark_root:
            self._roots.append(atom.index)
        self._atoms.append(atom)
        self._adj_list.append(list())
        self._bond_counts.append(0)
        self._ring_bond_flags.append(False)
        if atom.is_aromatic:
            self._delocal_subgraph[atom.index] = list()
        return atom

    def add_attribution(
            self,
            o: Union[DirectedBond, Atom],
            attr: List[Attribution]
    ) -> None:
        if self._attributable:
            if o in self._attribution:
                self._attribution[o].extend(attr)
            else:
                self._attribution[o] = attr

    def add_bond(
            self, src: int, dst: int,
            order: Union[int, float], stereo: str
    ) -> DirectedBond:
        assert src < dst

        bond = DirectedBond(src, dst, order, stereo, False)
        self._add_bond_at_loc(bond, -1)
        self._bond_counts[src] += order
        self._bond_counts[dst] += order

        if order == 1.5:
            self._delocal_subgraph.setdefault(src, []).append(dst)
            self._delocal_subgraph.setdefault(dst, []).append(src)
        return bond

    def add_placeholder_bond(self, src: int) -> int:
        out_edges = self._adj_list[src]
        out_edges.append(None)
        return len(out_edges) - 1

    def add_ring_bond(
            self, a: int, b: int,
            order: Union[int, float],
            a_stereo: Optional[str], b_stereo: Optional[str],
            a_pos: int = -1, b_pos: int = -1
    ) -> None:
        a_bond = DirectedBond(a, b, order, a_stereo, True)
        b_bond = DirectedBond(b, a, order, b_stereo, True)
        self._add_bond_at_loc(a_bond, a_pos)
        self._add_bond_at_loc(b_bond, b_pos)
        self._bond_counts[a] += order
        self._bond_counts[b] += order
        self._ring_bond_flags[a] = True
        self._ring_bond_flags[b] = True

        if order == 1.5:
            self._delocal_subgraph.setdefault(a, []).append(b)
            self._delocal_subgraph.setdefault(b, []).append(a)

    def update_bond_order(
            self, a: int, b: int,
            new_order: Union[int, float]
    ) -> None:
        assert 1 <= new_order <= 3

        if a > b:
            a, b = b, a  # swap so that a < b
        a_to_b = self._bond_dict[(a, b)]  # prev step guarantees existence
        if new_order == a_to_b.order:
            return
        elif a_to_b.ring_bond:
            b_to_a = self._bond_dict[(b, a)]
            bonds = (a_to_b, b_to_a)
        else:
            bonds = (a_to_b,)

        old_order = bonds[0].order
        for bond in bonds:
            bond.order = new_order
        self._bond_counts[a] += (new_order - old_order)
        self._bond_counts[b] += (new_order - old_order)

    def _add_bond_at_loc(self, bond, pos):
        self._bond_dict[(bond.src, bond.dst)] = bond

        out_edges = self._adj_list[bond.src]
        if (pos == -1) or (pos == len(out_edges)):
            out_edges.append(bond)
        elif out_edges[pos] is None:
            out_edges[pos] = bond
        else:
            out_edges.insert(pos, bond)

    def is_kekulized(self) -> bool:
        return not self._delocal_subgraph

    def kekulize(self) -> bool:
        # Algorithm based on Depth-First article by Richard L. Apodaca
        # Reference:
        #   https://depth-first.com/articles/2020/02/10/
        #   a-comprehensive-treatment-of-aromaticity-in-the-smiles-language/

        if self.is_kekulized():
            return True

        ds = self._delocal_subgraph
        kept_nodes = set(itertools.filterfalse(self._prune_from_ds, ds))

        # relabel kept DS nodes to be 0, 1, 2, ...
        label_to_node = list(sorted(kept_nodes))
        node_to_label = {v: i for i, v in enumerate(label_to_node)}

        # pruned and relabelled DS
        pruned_ds = [list() for _ in range(len(kept_nodes))]
        for node in kept_nodes:
            label = node_to_label[node]
            for adj in filter(lambda v: v in kept_nodes, ds[node]):
                pruned_ds[label].append(node_to_label[adj])

        matching = find_perfect_matching(pruned_ds)
        if matching is None:
            return False

        # de-aromatize and then make double bonds
        for node in ds:
            for adj in ds[node]:
                self.update_bond_order(node, adj, new_order=1)
            self._atoms[node].is_aromatic = False
            self._bond_counts[node] = int(self._bond_counts[node])

        for matched_labels in enumerate(matching):
            matched_nodes = tuple(label_to_node[i] for i in matched_labels)
            self.update_bond_order(*matched_nodes, new_order=2)

        self._delocal_subgraph = dict()  # clear DS
        return True

    def _prune_from_ds(self, node):
        adj_nodes = self._delocal_subgraph[node]
        if not adj_nodes:
            return True  # aromatic atom with no aromatic bonds

        atom = self._atoms[node]
        valences = AROMATIC_VALENCES[atom.element]

        # each bond in DS has order 1.5 - we treat them as single bonds
        used_electrons = int(self._bond_counts[node] - 0.5 * len(adj_nodes))

        if atom.h_count is None:  # account for implicit Hs
            assert atom.charge == 0
            return any(used_electrons == v for v in valences)
        else:
            valence = valences[-1] - atom.charge
            used_electrons += atom.h_count
            free_electrons = valence - used_electrons
            return not ((free_electrons >= 0) and (free_electrons % 2 != 0))



import enum
import re
from collections import deque
from typing import Iterator, Optional, Tuple, Union, List

from selfies.constants import AROMATIC_SUBSET, ELEMENTS, ORGANIC_SUBSET
from selfies.exceptions import SMILESParserError
from selfies.mol_graph import Atom, Attribution, \
    AttributionMap, DirectedBond, MolecularGraph

SMILES_BRACKETED_ATOM_PATTERN = re.compile(
    r"^[\[]"  # opening square bracket [
    r"(\d*)"  # isotope number (optional, e.g. 123, 26)
    r"([A-Za-z][a-z]?)"  # element symbol
    r"([@]{0,2})"  # chiral_tag (optional, only @ and @@ supported)
    r"((?:[H]\d?)?)"  # H count (optional, e.g. H, H0, H3)
    r"((?:[+]+|[-]+|[+-]\d+)?)"  # charge (optional, e.g. ---, +1, ++)
    r"((?:[:]\d+)?)"  # atom class (optional, e.g. :12, :1)
    r"[]]$"  # closing square bracket ]
)

SMILES_BOND_ORDERS = {"-": 1, "/": 1, "\\": 1, ":": 1.5, "=": 2, "#": 3}
SMILES_STEREO_BONDS = {"/", "\\"}


class SMILESTokenTypes(enum.Enum):
    ATOM = 0
    BRANCH = 1
    RING = 2
    DOT = 3


class SMILESToken:
    """A token in a SMILES string, containing a symbol (atom, branch bracket,
    ring number, dot) and its preceding bond, if it exists (e.g. =C, %12, #N).
    """

    def __init__(
            self,
            bond_idx: Optional[int],
            start_idx: int, end_idx: int, token_type: SMILESTokenTypes,
            token: str
    ):
        self.bond_idx = bond_idx
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.token_type = token_type
        self.token = token

    def extract_bond_char(self, smiles):
        return None if (self.bond_idx is None) else smiles[self.bond_idx]

    def extract_symbol(self, smiles):
        return smiles[self.start_idx:self.end_idx]

    def __str__(self):
        return self.token


def tokenize_smiles(smiles: str) -> Iterator[SMILESToken]:
    """Splits a SMILES string into its tokens.

    :param smiles: the input SMILES string.
    :return: the tokens of the input SMILES one-by-one with order preserved.
    """

    i = 0
    while i < len(smiles):

        if smiles[i] == ".":
            yield SMILESToken(None, i, i + 1, SMILESTokenTypes.DOT, smiles[i])
            i += 1
            continue

        if smiles[i] in SMILES_BOND_ORDERS:
            bond_idx = i
            i += 1
        else:
            bond_idx = None

        if i == len(smiles):
            raise SMILESParserError(smiles, "hanging bond", i - 1)

        elif smiles[i].isalpha():  # organic subset elements
            if smiles[i: i + 2] in ("Br", "Cl"):  # two-letter elements
                token = SMILESToken(bond_idx, i, i + 2,
                                    SMILESTokenTypes.ATOM, smiles[i: i + 2])
            else:  # one-letter elements (e.g. C, N, ...)
                token = SMILESToken(bond_idx, i, i + 1,
                                    SMILESTokenTypes.ATOM, smiles[i:i + 1])

        elif smiles[i] == "[":  # atoms encased in brackets (e.g. [NH])
            r_idx = smiles.find("]", i + 1)
            if r_idx == -1:
                raise SMILESParserError(smiles, "hanging bracket [", i)
            token = SMILESToken(bond_idx, i, r_idx + 1,
                                SMILESTokenTypes.ATOM, smiles[i:r_idx + 1])

        elif smiles[i] in ("(", ")"):  # open and closed branch brackets
            if bond_idx is not None:
                raise SMILESParserError(smiles, "hanging_bond", bond_idx)
            token = SMILESToken(
                None, i, i + 1, SMILESTokenTypes.BRANCH, smiles[i:i+1])

        elif smiles[i].isdigit():  # one-digit ring number
            token = SMILESToken(bond_idx, i, i + 1,
                                SMILESTokenTypes.RING, smiles[i:i+1])

        elif smiles[i] == "%":  # two-digit ring number (e.g. %12)
            rnum = smiles[i + 1: i + 3]
            if not (rnum.isnumeric() and len(rnum) == 2):
                err_msg = "invalid ring number '%{}'".format(rnum)
                raise SMILESParserError(smiles, err_msg, i)
            token = SMILESToken(bond_idx, i, i + 3,
                                SMILESTokenTypes.RING, smiles[i:i+3])

        else:
            err_msg = "unrecognized symbol '{}'".format(smiles[i])
            raise SMILESParserError(smiles, err_msg, i)

        yield token
        i = token.end_idx


# =============================================================================
# SMILES -> Atom, Graph, etc.
# =============================================================================


def smiles_to_atom(atom_symbol: str) -> Optional[Atom]:
    """Reads an atom from its SMILES representation.

    :param atom_symbol: a SMILES atom symbol.
    :return: the atom that the input symbol represents.
    """

    if atom_symbol[0] == "[" and atom_symbol[-1] == "]":
        pass  # continue below
    elif atom_symbol in ORGANIC_SUBSET:  # e.g. C, N, O, ...
        return Atom(atom_symbol, False)
    elif atom_symbol in AROMATIC_SUBSET:  # e.g. c, n, o, ...
        return Atom(atom_symbol.capitalize(), True)
    else:
        return None

    # e.g. [C], [C@@H], [O-], ...
    m = SMILES_BRACKETED_ATOM_PATTERN.match(atom_symbol)
    if m is None:
        return None
    isotope, element, chirality, h_count, charge, _ = m.groups()

    isotope = None if (isotope == "") else int(isotope)
    is_aromatic = element.islower() and (element in AROMATIC_SUBSET)
    element = element.capitalize()
    if element not in ELEMENTS:
        return None
    chirality = None if (chirality == "") else chirality

    s = h_count
    if s == "":
        h_count = 0
    else:
        s = s[1:]  # HXXX -> XXX
        h_count = 1 if (s == "") else int(s)

    s = charge
    if s == "":
        charge = 0
    else:
        if s[-1].isdigit():  # (+/-)XXX
            charge = int(s[1:])
        else:  # +++... or ---....
            charge = len(s)
        charge *= 1 if s[0] == "+" else -1

    return Atom(
        element=element,
        is_aromatic=is_aromatic,
        isotope=isotope,
        chirality=chirality,
        h_count=h_count,
        charge=charge
    )


def smiles_to_bond(
        bond_char: Optional[str]
) -> Tuple[Union[int, float], Optional[str]]:
    """Reads a bond from its SMILES representation.

    :param bond_char: a SMILES bond symbol.
    :return: the order and stereochemical specification of the bond
        that the input symbol represents.
    """

    order = SMILES_BOND_ORDERS.get(bond_char, 1)
    stereo = bond_char if (bond_char in SMILES_STEREO_BONDS) else None
    return order, stereo


def smiles_to_mol(smiles: str, attributable: bool) -> MolecularGraph:
    """Reads a molecular graph from a SMILES string.

    :param smiles: the input SMILES string.
    :param attributable: if molecular graph needs to include attributions
    :return: a molecular graph that the input SMILES string represents.
    :raises SMILESParserError: if the input SMILES is invalid.
    """

    if smiles == "":
        raise SMILESParserError(smiles, "empty SMILES", 0)

    mol = MolecularGraph(attributable=attributable)
    tokens = deque(tokenize_smiles(smiles))
    i = 0
    while tokens:
        i = _derive_mol_from_tokens(mol, smiles, tokens, i)
    return mol


def _derive_mol_from_tokens(mol, smiles, tokens, i):
    tok = None
    prev_stack = deque()  # keep track of previous atom on the current chain
    branch_stack = deque()  # keep track of open branches
    ring_log = dict()  # keep track of hanging ring numbers
    chain_start = True

    prev_stack.append(tok)
    while tokens:
        tok = tokens.popleft()
        bond_char = tok.extract_bond_char(smiles)
        symbol, symbol_type = tok.extract_symbol(smiles), tok.token_type
        prev_atom = prev_stack[-1]

        if symbol_type == SMILESTokenTypes.DOT:
            break

        elif symbol_type == SMILESTokenTypes.ATOM:
            curr = smiles_to_atom(symbol)
            if curr is None:
                err_msg = "invalid atom symbol '{}'".format(symbol)
                raise SMILESParserError(smiles, err_msg, tok.start_idx)

            curr, i = _attach_atom(mol, bond_char, curr, prev_atom, i, tok)
            prev_stack.pop()
            prev_stack.append(curr)
            chain_start = False

        elif chain_start:
            err_msg = "SMILES chain begins with non-atom"
            raise SMILESParserError(smiles, err_msg, tok.start_idx)

        elif symbol_type == SMILESTokenTypes.BRANCH:
            if symbol == "(":
                branch_stack.append(tok)
                prev_stack.append(prev_atom)
                chain_start = True
            else:
                if not branch_stack:
                    err_msg = "hanging ')' bracket"
                    raise SMILESParserError(smiles, err_msg, tok.start_idx)
                branch_stack.pop()
                prev_stack.pop()

        elif symbol_type == SMILESTokenTypes.RING:
            if symbol not in ring_log:
                lpos = mol.add_placeholder_bond(src=prev_atom.index)
                ring_log[symbol] = (tok, prev_atom, lpos)
            else:
                ltoken, latom, lpos = ring_log.pop(symbol)
                _make_ring_bonds(
                    mol=mol, smiles=smiles,
                    ltoken=ltoken, latom=latom, lpos=lpos,
                    rtoken=tok, ratom=prev_atom
                )

        else:
            # should not happen
            raise Exception("invalid symbol type")
        i += 1

    if len(mol) == 0:
        err_idx = (len(smiles) if (tok is None) else tok.start_idx) - 1
        raise SMILESParserError(smiles, "empty SMILES fragment", err_idx)

    if branch_stack:
        err_idx = branch_stack[-1].start_idx
        raise SMILESParserError(smiles, "hanging '(' bracket", err_idx)

    if ring_log:
        rnum, (tok, _, _) = list(ring_log.items())[-1]
        err_msg = "hanging ring number '{}'".format(rnum)
        raise SMILESParserError(smiles, err_msg, tok.start_idx)
    return i


def _attach_atom(mol, bond_char, atom, prev_atom, i, tok):
    is_root = (prev_atom is None)
    if bond_char:
        i += 1
    o = mol.add_atom(atom, mark_root=is_root)
    mol.add_attribution(o, [Attribution(i, str(tok))])
    if not is_root:
        src, dst = prev_atom.index, atom.index
        order, stereo = smiles_to_bond(bond_char)
        if prev_atom.is_aromatic and atom.is_aromatic and (bond_char is None):
            order = 1.5  # handle implicit aromatic bonds, e.g. cc
        o = mol.add_bond(src=src, dst=dst, order=order, stereo=stereo)
        mol.add_attribution(o, [Attribution(i, str(tok))])
    return atom, i


def _make_ring_bonds(mol, smiles, ltoken, latom, lpos, rtoken, ratom):
    if mol.has_bond(latom.index, ratom.index):
        err_msg = "ring bond specified between already-bonded atoms"
        raise SMILESParserError(smiles, err_msg, ltoken.start_idx)

    lbond_char = ltoken.extract_bond_char(smiles)
    rbond_char = rtoken.extract_bond_char(smiles)

    # checking that ring bonds match
    bonds = (lbond_char, rbond_char)
    if bonds[0] is None:
        bonds = (bonds[1], bonds[0])
    # swap bonds so that if bonds[0] is None, then bonds[1] is None

    if ((bonds[0] == bonds[1])
            or (bonds[1] is None)
            or all(x in SMILES_STEREO_BONDS for x in bonds)):
        pass
    else:
        err_msg = "mismatched ring bonds"
        raise SMILESParserError(smiles, err_msg, ltoken.start_idx)

    lorder, lstereo = smiles_to_bond(lbond_char)
    rorder, rstereo = smiles_to_bond(rbond_char)
    if latom.is_aromatic and ratom.is_aromatic and (bonds == (None, None)):
        lorder = rorder = 1.5  # handle implicit aromatic bonds, e.g. c1ccccc1

    mol.add_ring_bond(
        a=latom.index, a_stereo=lstereo, a_pos=lpos,
        b=ratom.index, b_stereo=rstereo,
        order=max(lorder, rorder)
    )


# =============================================================================
# SMILES <- Atom, Graph, etc.
# =============================================================================


def atom_to_smiles(atom: Atom, brackets: bool = True) -> str:
    """Converts an atom into its SMILES representation.

    :param atom: the input atom.
    :param brackets: True, if brackets should be added around the returned
        symbol (e.g. in the case of [C] or [C@@H]). Defaults to True.
    :return: a SMILES symbol representing the input atom.
    """
    assert not atom.is_aromatic

    specs = (atom.isotope, atom.chirality, atom.h_count, atom.charge)
    if specs == (None, None, None, 0):
        return atom.element
    else:
        builder = []
        if brackets:
            builder.append("[")
        if atom.isotope is not None:
            builder.append(str(atom.isotope))
        builder.append(atom.element)
        if atom.chirality is not None:
            builder.append(atom.chirality)
        if atom.h_count != 0:
            builder.append("H")
            builder.append(str(atom.h_count))
        elif specs == (None, None, 0, 0) and (atom.element in ORGANIC_SUBSET):
            builder.append("H0")
        if atom.charge != 0:
            builder.append("{:+}".format(atom.charge))
        if brackets:
            builder.append("]")

        return "".join(builder)


def bond_to_smiles(bond: DirectedBond) -> str:
    """Converts a bond into its SMILES representation.

    :param bond: the input bond.
    :return: a SMILES symbol representing the input bond.
    """

    if bond.order == 1:
        return bond.stereo if (bond.stereo in SMILES_STEREO_BONDS) else ""
    elif bond.order == 2:
        return "="
    elif bond.order == 3:
        return "#"
    else:  # this should never happen
        raise ValueError()


def mol_to_smiles(
    mol: MolecularGraph,
    attribute: bool = False
) -> Union[str, Tuple[str, List[Tuple[str,  List[Tuple[int, str]]]]]]:
    """Converts a molecular graph into its SMILES representation, maintaining
    the traversal order indicated by the input graph.

    :param mol: the input molecule.
    :param attribute: if an attribution should be returned
    :return: a SMILES string representing the input molecule if
             attribute is ``False``, otherwise a tuple is returned of
             SMILES string and attribution list.
    """
    assert mol.is_kekulized()

    fragments = []
    attribution_maps = []
    attribution_index = 0
    ring_log = dict()
    for root in mol.get_roots():
        derived = []
        _derive_smiles_from_fragment(
            derived, mol, root, ring_log, attribution_maps, attribution_index)
        attribution_index += _strlen(derived)
        fragments.append("".join(derived))
    # trim attribution map of empty tokens
    attribution_maps = [a for a in attribution_maps if a.token]
    result = ".".join(fragments), attribution_maps
    return result if attribute else result[0]


def _strlen(slist: List[str]) -> int:
    return len(''.join(slist))


def _derive_smiles_from_fragment(
        derived,
        mol,
        root,
        ring_log,
        attribution_maps, attribution_index=0):
    curr_atom, curr = mol.get_atom(root), root
    token = atom_to_smiles(curr_atom)
    derived.append(token)
    attribution_maps.append(AttributionMap(
        _strlen(derived) - 1 + attribution_index,
        token, mol.get_attribution(curr_atom)))

    out_bonds = mol.get_out_dirbonds(curr)
    for i, bond in enumerate(out_bonds):
        if bond.ring_bond:
            token = bond_to_smiles(bond)
            derived.append(token)
            attribution_maps.append(AttributionMap(
                _strlen(derived) - 1 + attribution_index,
                token, mol.get_attribution(bond)))
            ends = (min(bond.src, bond.dst), max(bond.src, bond.dst))
            rnum = ring_log.setdefault(ends, len(ring_log) + 1)
            if rnum >= 10:
                derived.append("%")
            derived.append(str(rnum))

        else:
            if i < len(out_bonds) - 1:
                derived.append("(")

            token = bond_to_smiles(bond)
            derived.append(token)
            attribution_maps.append(AttributionMap(
                _strlen(derived) - 1 + attribution_index,
                token, mol.get_attribution(bond)))
            _derive_smiles_from_fragment(
                derived, mol, bond.dst, ring_log,
                attribution_maps, attribution_index)
            if i < len(out_bonds) - 1:
                derived.append(")")
    return attribution_maps



from selfies.exceptions import EncoderError, SMILESParserError
from selfies.grammar_rules import get_selfies_from_index
from selfies.utils.smiles_utils import (
    atom_to_smiles,
    bond_to_smiles,
    smiles_to_mol
)

from selfies.mol_graph import AttributionMap


def encoder(smiles: str, strict: bool = True, attribute: bool = False) -> str:
    """Translates a SMILES string into its corresponding SELFIES string.

    This translation is deterministic and does not depend on the
    current semantic constraints. Additionally, it preserves the atom order
    of the input SMILES string; thus, one could generate randomized SELFIES
    strings by generating randomized SMILES strings, and then translating them.

    By nature of SELFIES, it is impossible to represent molecules that
    violate the current semantic constraints as SELFIES strings.
    Thus, we provide the ``strict`` flag to guard against such cases. If
    ``strict=True``, then this function will raise a
    :class:`selfies.EncoderError` if the input SMILES string represents
    a molecule that violates the semantic constraints. If
    ``strict=False``, then this function will not raise any error; however,
    calling :func:`selfies.decoder` on a SELFIES string generated this
    way will *not* be guaranteed to recover a SMILES string representing
    the original molecule.

    :param smiles: the SMILES string to be translated. It is recommended to
        use RDKit to check that the strings passed into this function
        are valid SMILES strings.
    :param strict: if ``True``, this function will check that the
        input SMILES string obeys the semantic constraints.
        Defaults to ``True``.
    :param attribute: if an attribution should be returned
    :return: a SELFIES string translated from the input SMILES string if
             attribute is ``False``, otherwise a tuple is returned of
             SELFIES string and attribution list.
    :raises EncoderError:  if the input SMILES string is invalid,
        cannot be kekulized, or violates the semantic constraints with
        ``strict=True``.

    :Example:

    >>> import selfies as sf
    >>> sf.encoder("C=CF")
    '[C][=C][F]'

    .. note:: This function does not currently support SMILES with:

        *   The wildcard symbol ``*``.
        *   The quadruple bond symbol ``$``.
        *   Chirality specifications other than ``@`` and ``@@``.
        *   Ring bonds across a dot symbol (e.g. ``c1cc([O-].[Na+])ccc1``) or
            ring bonds between atoms that are over 4000 atoms apart.

        Although SELFIES does not have aromatic symbols, this function
        *does* support aromatic SMILES strings by internally kekulizing them
        before translation.
    """

    try:
        mol = smiles_to_mol(smiles, attributable=attribute)
    except SMILESParserError as err:
        err_msg = "failed to parse input\n\tSMILES: {}".format(smiles)
        raise EncoderError(err_msg) from err

    if not mol.kekulize():
        err_msg = "kekulization failed\n\tSMILES: {}".format(smiles)
        raise EncoderError(err_msg)

    if strict:
        _check_bond_constraints(mol, smiles)

    # invert chirality of atoms where necessary,
    # such that they are restored when the SELFIES is decoded
    for atom in mol.get_atoms():
        if ((atom.chirality is not None)
                and mol.has_out_ring_bond(atom.index)
                and _should_invert_chirality(mol, atom)):
            atom.invert_chirality()

    fragments = []
    attribution_maps = []
    attribution_index = 0
    for root in mol.get_roots():
        derived = list(_fragment_to_selfies(
            mol, None, root, attribution_maps, attribution_index))
        attribution_index += len(derived)
        fragments.append("".join(derived))
    # trim attribution map of empty tokens
    attribution_maps = [a for a in attribution_maps if a.token]
    result = ".".join(fragments), attribution_maps
    return result if attribute else result[0]


def _check_bond_constraints(mol, smiles):
    errors = []

    for atom in mol.get_atoms():
        bond_cap = atom.bonding_capacity
        bond_count = mol.get_bond_count(atom.index)
        if bond_count > bond_cap:
            errors.append((atom_to_smiles(atom), bond_count, bond_cap))

    if errors:
        err_msg = "input violates the currently-set semantic constraints\n" \
                  "\tSMILES: {}\n" \
                  "\tErrors:\n".format(smiles)
        for e in errors:
            err_msg += "\t[{:} with {} bond(s) - " \
                       "a max. of {} bond(s) was specified]\n".format(*e)
        raise EncoderError(err_msg)


def _should_invert_chirality(mol, atom):
    out_bonds = mol.get_out_dirbonds(atom.index)

    # 1. rings whose right number are bonded to this atom (e.g. ...1...X1)
    # 2. rings whose left number are bonded to this atom (e.g. X1...1...)
    # 3. branches and other (e.g. X(...)...)
    partition = [[], [], []]
    for i, bond in enumerate(out_bonds):
        if not bond.ring_bond:
            partition[2].append(i)
        elif bond.src < bond.dst:
            partition[1].append(i)
        else:
            partition[0].append(i)
    partition[1].sort(key=lambda x: out_bonds[x].dst)

    # construct permutation
    perm = partition[0] + partition[1] + partition[2]
    count = 0
    for i in range(len(perm)):
        for j in range(i + 1, len(perm)):
            if perm[i] > perm[j]:
                count += 1
    return count % 2 != 0  # if odd permutation, should invert chirality


def _fragment_to_selfies(mol, bond_into_root, root,
                         attribution_maps, attribution_index=0):
    derived = []

    bond_into_curr, curr = bond_into_root, root
    while True:
        curr_atom = mol.get_atom(curr)
        token = _atom_to_selfies(bond_into_curr, curr_atom)
        derived.append(token)

        attribution_maps.append(AttributionMap(
            len(derived) - 1 + attribution_index,
            token, mol.get_attribution(curr_atom)))

        out_bonds = mol.get_out_dirbonds(curr)
        for i, bond in enumerate(out_bonds):

            if bond.ring_bond:
                if bond.src < bond.dst:
                    continue

                rev_bond = mol.get_dirbond(src=bond.dst, dst=bond.src)
                ring_len = bond.src - bond.dst
                Q_as_symbols = get_selfies_from_index(ring_len - 1)
                ring_symbol = "[{}Ring{}]".format(
                    _ring_bonds_to_selfies(rev_bond, bond),
                    len(Q_as_symbols)
                )

                derived.append(ring_symbol)
                attribution_maps.append(AttributionMap(
                    len(derived) - 1 + attribution_index,
                    ring_symbol, mol.get_attribution(bond)))
                for symbol in Q_as_symbols:
                    derived.append(symbol)
                    attribution_maps.append(AttributionMap(
                        len(derived) - 1 + attribution_index,
                        symbol, mol.get_attribution(bond)))

            elif i == len(out_bonds) - 1:
                bond_into_curr, curr = bond, bond.dst

            else:
                # start, end are so we can go back and
                # correct offset from branch symbol in
                # branch tokens
                start = len(attribution_maps)
                branch = _fragment_to_selfies(
                    mol, bond, bond.dst, attribution_maps, len(derived))
                Q_as_symbols = get_selfies_from_index(len(branch) - 1)
                branch_symbol = "[{}Branch{}]".format(
                    _bond_to_selfies(bond, show_stereo=False),
                    len(Q_as_symbols)
                )
                end = len(attribution_maps)

                derived.append(branch_symbol)
                for symbol in Q_as_symbols:
                    derived.append(symbol)
                    attribution_maps.append(AttributionMap(
                        len(derived) - 1 + attribution_index,
                        symbol, mol.get_attribution(bond)))

                # account for branch symbol because it is inserted after
                for j in range(start, end):
                    attribution_maps[j].index += len(Q_as_symbols) + 1
                attribution_maps.append(AttributionMap(
                    len(derived) - 1 + attribution_index,
                    branch_symbol, mol.get_attribution(bond)))

                derived.extend(branch)

        # end of chain
        if (not out_bonds) or out_bonds[-1].ring_bond:
            break
    return derived


def _bond_to_selfies(bond, show_stereo=True):
    if not show_stereo and (bond.order == 1):
        return ""
    return bond_to_smiles(bond)


def _ring_bonds_to_selfies(lbond, rbond):
    assert lbond.order == rbond.order

    if (lbond.order != 1) or all(b.stereo is None for b in (lbond, rbond)):
        return _bond_to_selfies(lbond, show_stereo=False)
    else:
        bond_char = "-" if (lbond.stereo is None) else lbond.stereo
        bond_char += "-" if (rbond.stereo is None) else rbond.stereo
        return bond_char


def _atom_to_selfies(bond, atom):
    assert not atom.is_aromatic
    bond_char = "" if (bond is None) else _bond_to_selfies(bond)
    return "[{}{}]".format(bond_char, atom_to_smiles(atom, brackets=False))



from typing import Dict, List, Tuple, Union

from selfies.utils.selfies_utils import len_selfies, split_selfies


def selfies_to_encoding(
        selfies: str,
        vocab_stoi: Dict[str, int],
        pad_to_len: int = -1,
        enc_type: str = 'both'
) -> Union[List[int], List[List[int]], Tuple[List[int], List[List[int]]]]:
    """Converts a SELFIES string into its label (integer)
    and/or one-hot encoding.

    A label encoded output will be a list of shape ``(L,)`` and a
    one-hot encoded output will be a 2D list of shape ``(L, len(vocab_stoi))``,
    where ``L`` is the symbol length of the SELFIES string. Optionally,
    the SELFIES string can be padded before it is encoded.

    :param selfies: the SELFIES string to be encoded.
    :param vocab_stoi: a dictionary that maps SELFIES symbols to indices,
        which must be non-negative and contiguous, starting from 0.
        If the SELFIES string is to be padded, then the special padding symbol
        ``[nop]`` must also be a key in this dictionary.
    :param pad_to_len: the length that the SELFIES string string is padded to.
        If this value is less than or equal to the symbol length of the
        SELFIES string, then no padding is added. Defaults to ``-1``.
    :param enc_type: the type of encoding of the output:
        ``label`` or ``one_hot`` or ``both``.
        If this value is ``both``, then a tuple of the label and one-hot
        encodings is returned. Defaults to ``both``.
    :return: the label encoded and/or one-hot encoded SELFIES string.

    :Example:

    >>> import selfies as sf
    >>> sf.selfies_to_encoding("[C][F]", {"[C]": 0, "[F]": 1})
    ([0, 1], [[1, 0], [0, 1]])
    """

    # some error checking
    if enc_type not in ("label", "one_hot", "both"):
        raise ValueError("enc_type must be in ('label', 'one_hot', 'both')")

    # pad with [nop]
    if pad_to_len > len_selfies(selfies):
        selfies += "[nop]" * (pad_to_len - len_selfies(selfies))

    # integer encode
    char_list = split_selfies(selfies)
    integer_encoded = [vocab_stoi[char] for char in char_list]

    if enc_type == "label":
        return integer_encoded

    # one-hot encode
    one_hot_encoded = list()
    for index in integer_encoded:
        letter = [0] * len(vocab_stoi)
        letter[index] = 1
        one_hot_encoded.append(letter)

    if enc_type == "one_hot":
        return one_hot_encoded
    return integer_encoded, one_hot_encoded


def encoding_to_selfies(
        encoding: Union[List[int], List[List[int]]],
        vocab_itos: Dict[int, str],
        enc_type: str,
) -> str:
    """Converts a label (integer) or one-hot encoding into a SELFIES string.

    If the input is label encoded, then a list of shape ``(L,)`` is
    expected; and if the input is one-hot encoded, then a 2D list of
    shape ``(L, len(vocab_itos))`` is expected.

    :param encoding: a label or one-hot encoding.
    :param vocab_itos: a dictionary that maps indices to SELFIES symbols.
        The indices of this dictionary must be non-negative and contiguous,
        starting from 0.
    :param enc_type: the type of encoding of the input:
        ``label`` or ``one_hot``.
    :return: the SELFIES string represented by the input encoding.

    :Example:

    >>> import selfies as sf
    >>> one_hot = [[0, 1, 0], [0, 0, 1], [1, 0, 0]]
    >>> vocab_itos = {0: "[nop]", 1: "[C]", 2: "[F]"}
    >>> sf.encoding_to_selfies(one_hot, vocab_itos, enc_type="one_hot")
    '[C][F][nop]'
    """

    if enc_type not in ("label", "one_hot"):
        raise ValueError("enc_type must be in ('label', 'one_hot')")

    if enc_type == "one_hot":  # Get integer encoding
        integer_encoded = []
        for row in encoding:
            integer_encoded.append(row.index(1))
    else:
        integer_encoded = encoding

    # Integer encoding -> SELFIES
    char_list = [vocab_itos[i] for i in integer_encoded]
    selfies = "".join(char_list)

    return selfies


def batch_selfies_to_flat_hot(
        selfies_batch: List[str],
        vocab_stoi: Dict[str, int],
        pad_to_len: int = -1,
) -> List[List[int]]:
    """Converts a list of SELFIES strings into its list of flattened
    one-hot encodings.

    Each SELFIES string in the input list is one-hot encoded
    (and then flattened) using :func:`selfies.selfies_to_encoding`, with
    ``vocab_stoi`` and ``pad_to_len`` being passed in as arguments.

    :param selfies_batch: the list of SELFIES strings to be encoded.
    :param vocab_stoi: a dictionary that maps SELFIES symbols to indices.
    :param pad_to_len: the length that each SELFIES string in the input list
        is padded to. Defaults to ``-1``.
    :return: the flattened one-hot encodings of the input list.

    :Example:

    >>> import selfies as sf
    >>> batch = ["[C]", "[C][C]"]
    >>> vocab_stoi = {"[nop]": 0, "[C]": 1}
    >>> sf.batch_selfies_to_flat_hot(batch, vocab_stoi, 2)
    [[0, 1, 1, 0], [0, 1, 0, 1]]
    """

    hot_list = list()

    for selfies in selfies_batch:
        one_hot = selfies_to_encoding(selfies, vocab_stoi, pad_to_len,
                                      enc_type="one_hot")
        flattened = [elem for vec in one_hot for elem in vec]
        hot_list.append(flattened)

    return hot_list


def batch_flat_hot_to_selfies(
        one_hot_batch: List[List[int]],
        vocab_itos: Dict[int, str],
) -> List[str]:
    """Converts a list of flattened one-hot encodings into a list
    of SELFIES strings.

    Each encoding in the input list is unflattened and then decoded using
    :func:`selfies.encoding_to_selfies`, with ``vocab_itos`` being passed in
    as an argument.

    :param one_hot_batch: a list of flattened one-hot encodings. Each
        encoding must be a list of length divisible by ``len(vocab_itos)``.
    :param vocab_itos: a dictionary that maps indices to SELFIES symbols.
    :return: the list of SELFIES strings represented by the input encodings.

    :Example:

    >>> import selfies as sf
    >>> batch = [[0, 1, 1, 0], [0, 1, 0, 1]]
    >>> vocab_itos = {0: "[nop]", 1: "[C]"}
    >>> sf.batch_flat_hot_to_selfies(batch, vocab_itos)
    ['[C][nop]', '[C][C]']
    """

    selfies_list = []

    for flat_one_hot in one_hot_batch:

        # Reshape to an L x M array where each column represents an alphabet
        # entry and each row is a position in the selfies
        one_hot = []

        M = len(vocab_itos)
        if len(flat_one_hot) % M != 0:
            raise ValueError("size of vector in one_hot_batch not divisible "
                             "by the length of the vocabulary.")
        L = len(flat_one_hot) // M

        for i in range(L):
            one_hot.append(flat_one_hot[M * i: M * (i + 1)])

        selfies = encoding_to_selfies(one_hot, vocab_itos, enc_type="one_hot")
        selfies_list.append(selfies)

    return selfies_list



import warnings
from typing import List, Union, Tuple

from selfies.compatibility import modernize_symbol
from selfies.exceptions import DecoderError
from selfies.grammar_rules import (
    get_index_from_selfies,
    next_atom_state,
    next_branch_state,
    next_ring_state,
    process_atom_symbol,
    process_branch_symbol,
    process_ring_symbol
)
from selfies.mol_graph import MolecularGraph, Attribution
from selfies.utils.selfies_utils import split_selfies
from selfies.utils.smiles_utils import mol_to_smiles


def decoder(
        selfies: str,
        compatible: bool = False,
        attribute: bool = False) ->\
        Union[str, Tuple[str, List[Tuple[str,  List[Tuple[int, str]]]]]]:
    """Translates a SELFIES string into its corresponding SMILES string.

    This translation is deterministic but depends on the current semantic
    constraints. The output SMILES string is guaranteed to be syntatically
    correct and guaranteed to represent a molecule that obeys the
    semantic constraints.

    :param selfies: the SELFIES string to be translated.
    :param compatible: if ``True``, this function will accept SELFIES strings
        containing depreciated symbols from previous releases. However, this
        function may behave differently than in previous major relases,
        and should not be treated as backard compatible.
        Defaults to ``False``.
    :param attribute: if ``True``, an attribution map connecting selfies
        tokens to smiles tokens is output.
    :return: a SMILES string derived from the input SELFIES string.
    :raises DecoderError: if the input SELFIES string is malformed.

    :Example:

    >>> import selfies as sf
    >>> sf.decoder('[C][=C][F]')
    'C=CF'
    """

    if compatible:
        msg = "\nselfies.decoder() may behave differently than in previous " \
              "major releases. We recommend using SELFIES that are up to date."
        warnings.warn(msg, stacklevel=2)

    mol = MolecularGraph(attributable=attribute)

    rings = []
    attribution_index = 0
    for s in selfies.split("."):
        n = _derive_mol_from_symbols(
            symbol_iter=enumerate(_tokenize_selfies(s, compatible)),
            mol=mol,
            selfies=selfies,
            max_derive=float("inf"),
            init_state=0,
            root_atom=None,
            rings=rings,
            attribute_stack=[] if attribute else None,
            attribution_index=attribution_index
        )
        attribution_index += n
    _form_rings_bilocally(mol, rings)
    return mol_to_smiles(mol, attribute)


def _tokenize_selfies(selfies, compatible):
    if isinstance(selfies, str):
        symbol_iter = split_selfies(selfies)
    elif isinstance(selfies, list):
        symbol_iter = selfies
    else:
        raise ValueError()  # should not happen

    try:
        for symbol in symbol_iter:
            if symbol == "[nop]":
                continue
            if compatible:
                symbol = modernize_symbol(symbol)
            yield symbol
    except ValueError as err:
        raise DecoderError(str(err)) from None


def _derive_mol_from_symbols(
        symbol_iter, mol, selfies, max_derive,
        init_state, root_atom, rings, attribute_stack, attribution_index
):
    n_derived = 0
    state = init_state
    prev_atom = root_atom

    while (state is not None) and (n_derived < max_derive):

        try:  # retrieve next symbol
            index, symbol = next(symbol_iter)
            n_derived += 1
        except StopIteration:
            break

        # Case 1: Branch symbol (e.g. [Branch1])
        if "ch" == symbol[-4:-2]:

            output = process_branch_symbol(symbol)
            if output is None:
                _raise_decoder_error(selfies, symbol)
            btype, n = output

            if state <= 1:
                next_state = state
            else:
                binit_state, next_state = next_branch_state(btype, state)

                Q = _read_index_from_selfies(symbol_iter, n_symbols=n)
                n_derived += n + _derive_mol_from_symbols(
                    symbol_iter, mol, selfies, (Q + 1),
                    init_state=binit_state, root_atom=prev_atom, rings=rings,
                    attribute_stack=attribute_stack +
                    [Attribution(index + attribution_index, symbol)
                     ] if attribute_stack is not None else None,
                    attribution_index=attribution_index
                )

        # Case 2: Ring symbol (e.g. [Ring2])
        elif "ng" == symbol[-4:-2]:

            output = process_ring_symbol(symbol)
            if output is None:
                _raise_decoder_error(selfies, symbol)
            ring_type, n, stereo = output

            if state == 0:
                next_state = state
            else:
                ring_order, next_state = next_ring_state(ring_type, state)
                bond_info = (ring_order, stereo)

                Q = _read_index_from_selfies(symbol_iter, n_symbols=n)
                n_derived += n
                lidx = max(0, prev_atom.index - (Q + 1))
                rings.append((mol.get_atom(lidx), prev_atom, bond_info))

        # Case 3: [epsilon]
        elif "eps" in symbol:
            next_state = 0 if (state == 0) else None

        # Case 4: regular symbol (e.g. [N], [=C], [F])
        else:

            output = process_atom_symbol(symbol)
            if output is None:
                _raise_decoder_error(selfies, symbol)
            (bond_order, stereo), atom = output
            cap = atom.bonding_capacity

            bond_order, next_state = next_atom_state(bond_order, cap, state)
            if bond_order == 0:
                if state == 0:
                    o = mol.add_atom(atom, True)
                    mol.add_attribution(
                        o,  attribute_stack +
                        [Attribution(index + attribution_index, symbol)]
                        if attribute_stack is not None else None)
            else:
                o = mol.add_atom(atom)
                mol.add_attribution(
                    o, attribute_stack +
                    [Attribution(index + attribution_index, symbol)]
                    if attribute_stack is not None else None)
                src, dst = prev_atom.index, atom.index
                o = mol.add_bond(src=src, dst=dst,
                                 order=bond_order, stereo=stereo)
                mol.add_attribution(
                    o, attribute_stack +
                    [Attribution(index + attribution_index, symbol)]
                    if attribute_stack is not None else None)
            prev_atom = atom

        if next_state is None:
            break
        state = next_state

    while n_derived < max_derive:  # consume remaining tokens
        try:
            next(symbol_iter)
            n_derived += 1
        except StopIteration:
            break

    return n_derived


def _raise_decoder_error(selfies, invalid_symbol):
    err_msg = "invalid symbol '{}'\n\tSELFIES: {}".format(
        invalid_symbol, selfies
    )
    raise DecoderError(err_msg)


def _read_index_from_selfies(symbol_iter, n_symbols):
    index_symbols = []
    for _ in range(n_symbols):
        try:
            index_symbols.append(next(symbol_iter)[-1])
        except StopIteration:
            index_symbols.append(None)
    return get_index_from_selfies(*index_symbols)


def _form_rings_bilocally(mol, rings):
    rings_made = [0] * len(mol)

    for latom, ratom, bond_info in rings:
        lidx, ridx = latom.index, ratom.index

        if lidx == ridx:  # ring to the same atom forbidden
            continue

        order, (lstereo, rstereo) = bond_info
        lfree = latom.bonding_capacity - mol.get_bond_count(lidx)
        rfree = ratom.bonding_capacity - mol.get_bond_count(ridx)

        if lfree <= 0 or rfree <= 0:
            continue  # no room for ring bond
        order = min(order, lfree, rfree)

        if mol.has_bond(a=lidx, b=ridx):
            bond = mol.get_dirbond(src=lidx, dst=ridx)
            new_order = min(order + bond.order, 3)
            mol.update_bond_order(a=lidx, b=ridx, new_order=new_order)

        else:
            mol.add_ring_bond(
                a=lidx, a_stereo=lstereo, a_pos=rings_made[lidx],
                b=ridx, b_stereo=rstereo, b_pos=rings_made[ridx],
                order=order
            )
            rings_made[lidx] += 1
            rings_made[ridx] += 1



from typing import Iterable, Iterator, Set


def len_selfies(selfies: str) -> int:
    """Returns the number of symbols in a given SELFIES string.

    :param selfies: a SELFIES string.
    :return: the symbol length of the SELFIES string.

    :Example:

    >>> import selfies as sf
    >>> sf.len_selfies("[C][=C][F].[C]")
    5
    """

    return selfies.count("[") + selfies.count(".")


def split_selfies(selfies: str) -> Iterator[str]:
    """Tokenizes a SELFIES string into its individual symbols.

    :param selfies: a SELFIES string.
    :return: the symbols of the SELFIES string one-by-one with order preserved.

    :Example:

    >>> import selfies as sf
    >>> list(sf.split_selfies("[C][=C][F].[C]"))
    ['[C]', '[=C]', '[F]', '.', '[C]']
    """

    left_idx = selfies.find("[")

    while 0 <= left_idx < len(selfies):
        right_idx = selfies.find("]", left_idx + 1)
        if right_idx == -1:
            raise ValueError("malformed SELFIES string, hanging '[' bracket")

        next_symbol = selfies[left_idx: right_idx + 1]
        yield next_symbol

        left_idx = right_idx + 1
        if selfies[left_idx: left_idx + 1] == ".":
            yield "."
            left_idx += 1


def get_alphabet_from_selfies(selfies_iter: Iterable[str]) -> Set[str]:
    """Constructs an alphabet from an iterable of SELFIES strings.

    The returned alphabet is the set of all symbols that appear in the
    SELFIES strings from the input iterable, minus the dot ``.`` symbol.

    :param selfies_iter: an iterable of SELFIES strings.
    :return: an alphabet of SELFIES symbols, built from the input iterable.

    :Example:

    >>> import selfies as sf
    >>> selfies_list = ["[C][F][O]", "[C].[O]", "[F][F]"]
    >>> alphabet = sf.get_alphabet_from_selfies(selfies_list)
    >>> sorted(list(alphabet))
    ['[C]', '[F]', '[O]']
    """

    alphabet = set()
    for s in selfies_iter:
        for symbol in split_selfies(s):
            alphabet.add(symbol)
    alphabet.discard(".")
    return alphabet



class SMILESParserError(ValueError):
    """Exception raised when a SMILES fails to be parsed.
    """

    def __init__(self, smiles, reason="N/A", idx=-1):
        self.smiles = smiles
        self.idx = idx
        self.reason = reason

    def __str__(self):
        err_msg = "\n" \
                  "\tSMILES: {smiles}\n" \
                  "\t        {pointer}\n" \
                  "\tIndex:  {index}\n" \
                  "\tReason: {reason}"

        return err_msg.format(
            smiles=self.smiles,
            pointer=(" " * self.idx + "^"),
            index=self.idx,
            reason=self.reason
        )


class EncoderError(Exception):
    """Exception raised by :func:`selfies.encoder`.
    """

    pass


class DecoderError(Exception):
    """Exception raised by :func:`selfies.decoder`.
    """

    pass



import heapq
import itertools
from collections import deque
from typing import List, Optional


def find_perfect_matching(graph: List[List[int]]) -> Optional[List[int]]:
    """Finds a perfect matching for an undirected graph (without self-loops).

    :param graph: an adjacency list representing the input graph.
    :return: a list representing a perfect matching, where j is the i-th
        element if nodes i and j are matched. Returns None, if the graph cannot
        be perfectly matched.
    """

    # start with a maximal matching for efficiency
    matching = _greedy_matching(graph)

    unmatched = set(i for i in range(len(graph)) if matching[i] is None)
    while unmatched:

        # find augmenting path which starts at root
        root = unmatched.pop()
        path = _find_augmenting_path(graph, root, matching)

        if path is None:
            return None
        else:
            _flip_augmenting_path(matching, path)
            unmatched.discard(path[0])
            unmatched.discard(path[-1])

    return matching


def _greedy_matching(graph):
    matching = [None] * len(graph)
    free_degrees = [len(graph[i]) for i in range(len(graph))]
    # free_degrees[i] = number of unmatched neighbors for node i

    # prioritize nodes with fewer unmatched neighbors
    node_pqueue = [(free_degrees[i], i) for i in range(len(graph))]
    heapq.heapify(node_pqueue)

    while node_pqueue:
        _, node = heapq.heappop(node_pqueue)

        if (matching[node] is not None) or (free_degrees[node] == 0):
            continue  # node cannot be matched

        # match node with first unmatched neighbor
        mate = next(i for i in graph[node] if matching[i] is None)
        matching[node] = mate
        matching[mate] = node

        for adj in itertools.chain(graph[node], graph[mate]):
            free_degrees[adj] -= 1
            if (matching[adj] is None) and (free_degrees[adj] > 0):
                heapq.heappush(node_pqueue, (free_degrees[adj], adj))

    return matching


def _find_augmenting_path(graph, root, matching):
    assert matching[root] is None

    # run modified BFS to find path from root to unmatched node
    other_end = None
    node_queue = deque([root])

    # parent BFS tree - None indicates an unvisited node
    parents = [None] * len(graph)
    parents[root] = [None, None]

    while node_queue:
        node = node_queue.popleft()

        for adj in graph[node]:
            if matching[adj] is None:  # unmatched node
                if adj != root:  # augmenting path found!
                    parents[adj] = [node, adj]
                    other_end = adj
                    break
            else:
                adj_mate = matching[adj]
                if parents[adj_mate] is None:  # adj_mate not visited
                    parents[adj_mate] = [node, adj]
                    node_queue.append(adj_mate)

        if other_end is not None:
            break  # augmenting path found!

    if other_end is None:
        return None
    else:
        path = []
        node = other_end
        while node != root:
            path.append(parents[node][1])
            path.append(parents[node][0])
            node = parents[node][0]
        return path


def _flip_augmenting_path(matching, path):
    for i in range(0, len(path), 2):
        a, b = path[i], path[i + 1]
        matching[a] = b
        matching[b] = a
